# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from logging import getLogger, DEBUG as _DEBUG
_log = getLogger(__name__)
del getLogger

from gi.repository import Gtk
import solaar.gtk as gtk

#
#
#

_LARGE_SIZE = 64
Gtk.IconSize.LARGE = Gtk.icon_size_register('large', _LARGE_SIZE, _LARGE_SIZE)
# Gtk.IconSize.XLARGE = Gtk.icon_size_register('x-large', _LARGE_SIZE * 2, _LARGE_SIZE * 2)
# print ("menu", int(Gtk.IconSize.MENU), Gtk.icon_size_lookup(Gtk.IconSize.MENU))
# print ("small toolbar", int(Gtk.IconSize.SMALL_TOOLBAR), Gtk.icon_size_lookup(Gtk.IconSize.SMALL_TOOLBAR))
# print ("button", int(Gtk.IconSize.BUTTON), Gtk.icon_size_lookup(Gtk.IconSize.BUTTON))
# print ("large toolbar", int(Gtk.IconSize.LARGE_TOOLBAR), Gtk.icon_size_lookup(Gtk.IconSize.LARGE_TOOLBAR))
# print ("dnd", int(Gtk.IconSize.DND), Gtk.icon_size_lookup(Gtk.IconSize.DND))
# print ("dialog", int(Gtk.IconSize.DIALOG), Gtk.icon_size_lookup(Gtk.IconSize.DIALOG))

TRAY_INIT = 'solaar-init'
TRAY_OKAY = 'solaar'
TRAY_ATTENTION = 'solaar-attention'


def _look_for_application_icons():
	import os.path as _path
	from os import environ as _environ

	import sys as _sys
	if _log.isEnabledFor(_DEBUG):
		_log.debug("sys.path[0] = %s", _sys.path[0])
	prefix_share = _path.normpath(_path.join(_path.realpath(_sys.path[0]), '..'))
	src_share = _path.normpath(_path.join(_path.realpath(_sys.path[0]), '..', 'share'))
	local_share = _environ.get('XDG_DATA_HOME', _path.expanduser(_path.join('~', '.local', 'share')))
	data_dirs = _environ.get('XDG_DATA_DIRS', '/usr/local/share:/usr/share')
	repo_share = _path.normpath(_path.join(_path.dirname(__file__), '..', '..', '..', 'share'))
	setuptools_share = _path.normpath(_path.join(_path.dirname(__file__), '..', '..', 'share'))
	del _sys

	share_solaar = [prefix_share] + list(_path.join(x, 'solaar') for x in [src_share, local_share, setuptools_share, repo_share] + data_dirs.split(':'))
	for location in share_solaar:
		location = _path.join(location, 'icons')
		if _log.isEnabledFor(_DEBUG):
			_log.debug("looking for icons in %s", location)

		if _path.exists(_path.join(location, TRAY_ATTENTION + '.svg')):
			yield location

	del _environ
	# del _path


_default_theme = None
_use_symbolic_icons = False

def _init_icon_paths():
	global _default_theme
	if _default_theme:
		return

	_default_theme = Gtk.IconTheme.get_default()
	for p in _look_for_application_icons():
		_default_theme.prepend_search_path(p)
	if _log.isEnabledFor(_DEBUG):
		_log.debug("icon theme paths: %s", _default_theme.get_search_path())

	if gtk.prefer_symbolic_battery_icons:
		if _default_theme.has_icon('battery-good-symbolic'):
			global _use_symbolic_icons
			_use_symbolic_icons = True
			return
		else:
			_log.warning("failed to detect symbolic icons")
	if not _default_theme.has_icon('battery-good'):
		_log.warning("failed to detect icons")

#
#
#

def battery(level=None, charging=False):
	icon_name = _battery_icon_name(level, charging)
	if not _default_theme.has_icon(icon_name):
		_log.warning("icon %s not found in current theme", icon_name)
		return TRAY_OKAY # use Solaar icon if battery icon not available
	elif _log.isEnabledFor(_DEBUG):
		_log.debug("battery icon for %s:%s = %s", level, charging, icon_name)
	return icon_name

# return first res where val >= guard
# _first_res(val,((guard,res),...))
def _first_res(val,pairs):
	return next((res for guard,res in pairs if val >= guard),None)

def _battery_icon_name(level, charging):
	_init_icon_paths()

	if level is None or level < 0:
		return 'battery-missing' + ( '-symbolic' if _use_symbolic_icons else '' )

	level_name = _first_res(level,((90,'full'), (50,'good'), (20,'low'), (5,'caution'), (0,'empty')))
	return 'battery-%s%s%s' % (level_name, '-charging' if charging else '', '-symbolic' if _use_symbolic_icons else '')

#
#
#

def lux(level=None):
	if level is None or level < 0:
		return 'light_unknown'
	return 'light_%03d' % (20 * ((level + 50) // 100))

#
#
#

_ICON_SETS = {}

def device_icon_set(name='_', kind=None):
	icon_set = _ICON_SETS.get(name)
	if icon_set is None:
		icon_set = Gtk.IconSet.new()
		_ICON_SETS[name] = icon_set

		# names of possible icons, in reverse order of likelihood
		# the theme will hopefully pick up the most appropriate
		names = ['preferences-desktop-peripherals']
		if kind:
			if str(kind) == 'numpad':
				names += ('input-keyboard', 'input-dialpad')
			elif str(kind) == 'touchpad':
				names += ('input-mouse', 'input-tablet')
			elif str(kind) == 'trackball':
				names += ('input-mouse',)
			names += ('input-' + str(kind),)
		# names += (name.replace(' ', '-'),)

		source = Gtk.IconSource.new()
		for n in names:
			source.set_icon_name(n)
			icon_set.add_source(source)
		icon_set.names = names

	return icon_set


def device_icon_file(name, kind=None, size=_LARGE_SIZE):
	_init_icon_paths()

	icon_set = device_icon_set(name, kind)
	assert icon_set
	for n in reversed(icon_set.names):
		if _default_theme.has_icon(n):
			return _default_theme.lookup_icon(n, size, 0).get_filename()


def device_icon_name(name, kind=None):
	_init_icon_paths()

	icon_set = device_icon_set(name, kind)
	assert icon_set
	for n in reversed(icon_set.names):
		if _default_theme.has_icon(n):
			return n


def icon_file(name, size=_LARGE_SIZE):
	_init_icon_paths()

        # has_icon() somehow returned False while lookup_icon returns non-None.
        # I guess it happens because share/solaar/icons/ has no hicolor and
        # resolution subdirs
	theme_icon = _default_theme.lookup_icon(name, size, 0)
	if theme_icon:
		file_name = theme_icon.get_filename()
		# if _log.isEnabledFor(_DEBUG):
		# 	_log.debug("icon %s(%d) => %s", name, size, file_name)
		return file_name

	_log.warn("icon %s(%d) not found in current theme", name, size)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2020
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals


from time import time as _timestamp

from logitech_receiver.common import strhex as _strhex

from logitech_receiver import (
				base as _base,
				hidpp10 as _hidpp10,
				status as _status,
				notifications as _notifications,
			)

_R = _hidpp10.REGISTERS

from solaar.cli.show import _print_receiver

def run(receivers, args, find_receiver, _ignore):
	assert receivers

	if args.receiver:
		receiver_name = args.receiver.lower()
		receiver = find_receiver(receiver_name)
		if not receiver:
			raise Exception("no receiver found matching '%s'" % receiver_name)
	else:
		receiver = receivers[0]

	assert receiver

	_print_receiver(receiver)

	print ('  Register Dump')
	register = receiver.read_register(_R.notifications)
	print("    Notification Register %#04x: %s" % (_R.notifications%0x100,'0x'+_strhex(register) if register else "None"))
	register = receiver.read_register(_R.receiver_connection)
	print("    Connection State      %#04x: %s" % (_R.receiver_connection%0x100,'0x'+_strhex(register) if register else "None"))
	register = receiver.read_register(_R.devices_activity)
	print("    Device Activity       %#04x: %s" % (_R.devices_activity%0x100,'0x'+_strhex(register) if register else "None"))

	for device in range(0,6):
		for sub_reg in [ 0x0, 0x10, 0x20, 0x30 ] :
			register = receiver.read_register(_R.receiver_info, sub_reg + device)
			print("    Pairing Register %#04x %#04x: %s" % (_R.receiver_info%0x100,sub_reg + device,'0x'+_strhex(register) if register else "None"))
		register = receiver.read_register(_R.receiver_info, 0x40 + device)
		print("    Pairing Name     %#04x %#02x: %s" % (_R.receiver_info%0x100,0x40 + device,register[2:2+ord(register[1:2])] if register else "None"))

	for sub_reg in range(0,5):
		register = receiver.read_register(_R.firmware, sub_reg)
		print("    Firmware         %#04x %#04x: %s" % (_R.firmware%0x100,sub_reg,'0x'+_strhex(register) if register else "None"))



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Optional desktop notifications.

from __future__ import absolute_import, division, print_function, unicode_literals


from solaar.i18n import _

#
#
#

try:
	import gi
	gi.require_version('Notify', '0.7')
	# this import is allowed to fail, in which case the entire feature is unavailable
	from gi.repository import Notify, GLib

	# assumed to be working since the import succeeded
	available = True

except (ValueError, ImportError):
	available = False

if available:
	from logging import getLogger, INFO as _INFO
	_log = getLogger(__name__)
	del getLogger

	from solaar import NAME
	from . import icons as _icons

	# cache references to shown notifications here, so if another status comes
	# while its notification is still visible we don't create another one
	_notifications = {}

	def init():
		"""Init the notifications system."""
		global available
		if available:
			if not Notify.is_initted():
				if _log.isEnabledFor(_INFO):
					_log.info("starting desktop notifications")
				try:
					return Notify.init(NAME)
				except:
					_log.exception("initializing desktop notifications")
					available = False
		return available and Notify.is_initted()


	def uninit():
		if available and Notify.is_initted():
			if _log.isEnabledFor(_INFO):
				_log.info("stopping desktop notifications")
			_notifications.clear()
			Notify.uninit()


	# def toggle(action):
	# 	if action.get_active():
	# 		init()
	# 	else:
	# 		uninit()
	# 	action.set_sensitive(available)
	# 	return action.get_active()


	def alert(reason, icon=None):
		assert reason

		if available and Notify.is_initted():
			n = _notifications.get(NAME)
			if n is None:
				n = _notifications[NAME] = Notify.Notification()

			# we need to use the filename here because the notifications daemon
			# is an external application that does not know about our icon sets
			icon_file = _icons.icon_file(NAME.lower()) if icon is None \
						else _icons.icon_file(icon)

			n.update(NAME, reason, icon_file)
			n.set_urgency(Notify.Urgency.NORMAL)
			n.set_hint("desktop-entry", GLib.Variant('s', NAME.lower()))

			try:
				# if _log.isEnabledFor(_DEBUG):
				# 	_log.debug("showing %s", n)
				n.show()
			except Exception:
				_log.exception("showing %s", n)


	def show(dev, reason=None, icon=None):
		"""Show a notification with title and text."""
		if available and Notify.is_initted():
			summary = dev.name

			# if a notification with same name is already visible, reuse it to avoid spamming
			n = _notifications.get(summary)
			if n is None:
				n = _notifications[summary] = Notify.Notification()

			if reason:
				message = reason
			elif dev.status is None:
				message = _("unpaired")
			elif bool(dev.status):
				message = dev.status.to_string() or _("connected")
			else:
				message = _("offline")

			# we need to use the filename here because the notifications daemon
			# is an external application that does not know about our icon sets
			icon_file = _icons.device_icon_file(dev.name, dev.kind) if icon is None \
						else _icons.icon_file(icon)

			n.update(summary, message, icon_file)
			urgency = Notify.Urgency.LOW if dev.status else Notify.Urgency.NORMAL
			n.set_urgency(urgency)
			n.set_hint("desktop-entry", GLib.Variant('s', NAME.lower()))

			try:
				# if _log.isEnabledFor(_DEBUG):
				# 	_log.debug("showing %s", n)
				n.show()
			except Exception:
				_log.exception("showing %s", n)

else:
	init = lambda: False
	uninit = lambda: None
	# toggle = lambda action: False
	alert = lambda reason: None
	show = lambda dev, reason=None: None



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# USB ids of Logitech wireless receivers.
# Only receivers supporting the HID++ protocol can go in here.

from __future__ import absolute_import, division, print_function, unicode_literals


_DRIVER = ('hid-generic', 'generic-usb', 'logitech-djreceiver')

# max_devices is only used for receivers that do not support reading from _R.receiver_info offset 0x03, default to 1
# may_unpair is only used for receivers that do not support reading from _R.receiver_info offset 0x03, default to False
## should this last be changed so that may_unpair is used for all receivers? writing to _R.receiver_pairing doesn't seem right
# re_pairs determines whether a receiver pairs by replacing existing pairings, default to False
## currently only one receiver is so marked - should there be more?

_unifying_receiver = lambda product_id: {
	'vendor_id':0x046d,
	'product_id':product_id, 
	'usb_interface':2,
	'hid_driver':_DRIVER,
	'name':'Unifying Receiver'
}

_nano_receiver = lambda product_id: {
	'vendor_id':0x046d,
	'product_id':product_id,
	'usb_interface':1,
	'hid_driver':_DRIVER,
	'name':'Nano Receiver',
	'may_unpair': False,
	're_pairs': True 
}

_nano_receiver_max2 = lambda product_id: {
	'vendor_id':0x046d,
	'product_id':product_id,
	'usb_interface':1,
	'hid_driver':_DRIVER,
	'name':'Nano Receiver',
	'max_devices': 2,
	'may_unpair': False,
	're_pairs': True 
}

_nano_receiver_maxn = lambda product_id, max: {
	'vendor_id':0x046d,
	'product_id':product_id,
	'usb_interface':1,
	'hid_driver':_DRIVER,
	'name':'Nano Receiver',
	'max_devices': max,
	'may_unpair': False,
	're_pairs': True 
}

_lenovo_receiver = lambda product_id: {
	'vendor_id':0x17ef, 
	'product_id':product_id, 
	'usb_interface':1, 
	'hid_driver':_DRIVER, 
	'name':'Nano Receiver'
}

_lightspeed_receiver = lambda product_id: {
	'vendor_id':0x046d,
	'product_id':product_id,
	'usb_interface':2,
	'hid_driver':_DRIVER,
	'name':'Lightspeed Receiver'
}

# standard Unifying receivers (marked with the orange Unifying logo)
UNIFYING_RECEIVER_C52B    = _unifying_receiver(0xc52b)
UNIFYING_RECEIVER_C532    = _unifying_receiver(0xc532)

# Nano receviers that support the Unifying protocol
NANO_RECEIVER_ADVANCED    = _nano_receiver(0xc52f)

# Nano receivers that don't support the Unifying protocol
NANO_RECEIVER_C517        = _nano_receiver_maxn(0xc517,6)
NANO_RECEIVER_C518        = _nano_receiver(0xc518)
NANO_RECEIVER_C51A        = _nano_receiver(0xc51a)
NANO_RECEIVER_C51B        = _nano_receiver(0xc51b)
NANO_RECEIVER_C521        = _nano_receiver(0xc521)
NANO_RECEIVER_C525        = _nano_receiver(0xc525)
NANO_RECEIVER_C526        = _nano_receiver(0xc526)
NANO_RECEIVER_C52e        = _nano_receiver(0xc52e)
NANO_RECEIVER_C531        = _nano_receiver(0xc531)
NANO_RECEIVER_C534        = _nano_receiver_max2(0xc534)
NANO_RECEIVER_C537        = _nano_receiver(0xc537)
NANO_RECEIVER_6042        = _lenovo_receiver(0x6042)

# Lightspeed receivers
LIGHTSPEED_RECEIVER_C539  = _lightspeed_receiver(0xc539)
LIGHTSPEED_RECEIVER_C53a  = _lightspeed_receiver(0xc53a)
LIGHTSPEED_RECEIVER_C53f  = _lightspeed_receiver(0xc53f)

del _DRIVER, _unifying_receiver, _nano_receiver, _lenovo_receiver, _lightspeed_receiver


ALL = (
		UNIFYING_RECEIVER_C52B,
		UNIFYING_RECEIVER_C532,
		NANO_RECEIVER_ADVANCED,
		NANO_RECEIVER_C517,
		NANO_RECEIVER_C518,
		NANO_RECEIVER_C51A,
		NANO_RECEIVER_C51B,
		NANO_RECEIVER_C521,
		NANO_RECEIVER_C525,
		NANO_RECEIVER_C526,
		NANO_RECEIVER_C52e,
		NANO_RECEIVER_C531,
		NANO_RECEIVER_C534,
		NANO_RECEIVER_C537,
		NANO_RECEIVER_6042,
		LIGHTSPEED_RECEIVER_C539,
		LIGHTSPEED_RECEIVER_C53a,
		LIGHTSPEED_RECEIVER_C53f,
	)

def product_information(usb_id):
	if isinstance(usb_id,str):
		usb_id = int(usb_id,16)
	for r in ALL:
		if usb_id == r.get('product_id'):
			return r
	return { }



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from gi.repository import Gtk, GLib
from threading import Timer as _Timer

from solaar.i18n import _
from solaar.ui import ui_async as _ui_async
from logitech_receiver.settings import KIND as _SETTING_KIND

#
#
#

def _read_async(setting, force_read, sbox, device_is_online):
	def _do_read(s, force, sb, online):
		v = s.read(not force)
		GLib.idle_add(_update_setting_item, sb, v, online, priority=99)

	_ui_async(_do_read, setting, force_read, sbox, device_is_online)


def _write_async(setting, value, sbox):
	_ignore, failed, spinner, control = sbox.get_children()
	control.set_sensitive(False)
	failed.set_visible(False)
	spinner.set_visible(True)
	spinner.start()

	def _do_write(s, v, sb):
		v = setting.write(v)
		GLib.idle_add(_update_setting_item, sb, v, True, priority=99)

	_ui_async(_do_write, setting, value, sbox)

#
#
#

def _create_toggle_control(setting):
	def _switch_notify(switch, _ignore, s):
		if switch.get_sensitive():
			_write_async(s, switch.get_active() == True, switch.get_parent())

	c = Gtk.Switch()
	c.connect('notify::active', _switch_notify, setting)
	return c

def _create_choice_control(setting):
	def _combo_notify(cbbox, s):
		if cbbox.get_sensitive():
			_write_async(s, cbbox.get_active_id(), cbbox.get_parent())

	c = Gtk.ComboBoxText()
	# TODO i18n text entries
	for entry in setting.choices:
		c.append(str(entry), str(entry))
	c.connect('changed', _combo_notify, setting)
	return c

def _create_slider_control(setting):
	class SliderControl:
		__slots__ = ('gtk_range', 'timer', 'setting')
		def __init__(self, setting):
			self.setting = setting
			self.timer = None

			self.gtk_range = Gtk.Scale()
			self.gtk_range.set_range(*self.setting.range)
			self.gtk_range.set_round_digits(0)
			self.gtk_range.set_digits(0)
			self.gtk_range.set_increments(1, 5)
			self.gtk_range.connect('value-changed',
			                       lambda _, c: c._changed(),
								   self)

		def _write(self):
			_write_async(self.setting,
			             int(self.gtk_range.get_value()),
						 self.gtk_range.get_parent())
			self.timer.cancel()

		def _changed(self):
			if self.gtk_range.get_sensitive():
				if self.timer:
					self.timer.cancel()
				self.timer = _Timer(0.5, lambda: GLib.idle_add(self._write))
				self.timer.start()

	control = SliderControl(setting)
	return control.gtk_range

#
#
#

def _create_sbox(s):
	sbox = Gtk.HBox(homogeneous=False, spacing=6)
	sbox.pack_start(Gtk.Label(s.label), False, False, 0)

	spinner = Gtk.Spinner()
	spinner.set_tooltip_text(_("Working") + '...')

	failed = Gtk.Image.new_from_icon_name('dialog-warning', Gtk.IconSize.SMALL_TOOLBAR)
	failed.set_tooltip_text(_("Read/write operation failed."))

	if s.kind == _SETTING_KIND.toggle:
		control = _create_toggle_control(s)
		sbox.pack_end(control, False, False, 0)
	elif s.kind == _SETTING_KIND.choice:
		control = _create_choice_control(s)
		sbox.pack_end(control, False, False, 0)
	elif s.kind == _SETTING_KIND.range:
		control = _create_slider_control(s)
		sbox.pack_end(control, True, True, 0)
	else:
		raise NotImplemented

	control.set_sensitive(False)  # the first read will enable it
	sbox.pack_end(spinner, False, False, 0)
	sbox.pack_end(failed, False, False, 0)

	if s.description:
		sbox.set_tooltip_text(s.description)

	sbox.show_all()
	spinner.start()  # the first read will stop it
	failed.set_visible(False)

	return sbox


def _update_setting_item(sbox, value, is_online=True):
	_ignore, failed, spinner, control = sbox.get_children()
	spinner.set_visible(False)
	spinner.stop()

	# print ("update", control, "with new value", value)
	if value is None:
		control.set_sensitive(False)
		failed.set_visible(is_online)
		return

	failed.set_visible(False)
	if isinstance(control, Gtk.Switch):
		control.set_active(value)
	elif isinstance(control, Gtk.ComboBoxText):
		control.set_active_id(str(value))
	elif isinstance(control, Gtk.Scale):
		control.set_value(int(value))
	else:
		raise NotImplemented
	control.set_sensitive(True)

#
#
#

# config panel
_box = None
_items = {}

def create():
	global _box
	assert _box is None
	_box = Gtk.VBox(homogeneous=False, spacing=8)
	_box._last_device = None
	return _box


def update(device, is_online=None):
	assert _box is not None
	assert device
	device_id = (device.receiver.path, device.number)
	if is_online is None:
		is_online = bool(device.online)

	# if the device changed since last update, clear the box first
	if device_id != _box._last_device:
		_box.set_visible(False)
		_box._last_device = device_id

	# hide controls belonging to other devices
	for k, sbox in _items.items():
		sbox = _items[k]
		sbox.set_visible(k[0:2] == device_id)

	for s in device.settings:
		k = (device_id[0], device_id[1], s.name)
		if k in _items:
			sbox = _items[k]
		else:
			sbox = _items[k] = _create_sbox(s)
			_box.pack_start(sbox, False, False, 0)

		_read_async(s, False, sbox, is_online)

	_box.set_visible(True)


def clean(device):
	"""Remove the controls for a given device serial.
	Needed after the device has been unpaired.
	"""
	assert _box is not None
	device_id = (device.receiver.path, device.number)
	for k in list(_items.keys()):
		if k[0:2] == device_id:
			_box.remove(_items[k])
			del _items[k]


def destroy():
	global _box
	_box = None
	_items.clear()



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from logging import getLogger, DEBUG as _DEBUG
_log = getLogger(__name__)
del getLogger

from copy import copy as _copy
import math

from .common import (
				NamedInt as _NamedInt,
				NamedInts as _NamedInts,
				bytes2int as _bytes2int,
				int2bytes as _int2bytes,
			)

#
#
#

KIND = _NamedInts(toggle=0x01, choice=0x02, range=0x04)

class Setting(object):
	"""A setting descriptor.
	Needs to be instantiated for each specific device."""
	__slots__ = ('name', 'label', 'description', 'kind', 'device_kind',
					'_rw', '_validator', '_device', '_value')

	def __init__(self, name, rw, validator, kind=None, label=None, description=None, device_kind=None):
		assert name
		self.name = name
		self.label = label or name
		self.description = description
		self.device_kind = device_kind

		self._rw = rw
		self._validator = validator

		assert kind is None or kind & validator.kind != 0
		self.kind = kind or validator.kind

	def __call__(self, device):
		assert not hasattr(self, '_value')
		# combined keyboards and touchpads (e.g., K400) break this assertion so don't use it
		# assert self.device_kind is None or device.kind in self.device_kind
		p = device.protocol
		if p == 1.0:
			# HID++ 1.0 devices do not support features
			assert self._rw.kind == RegisterRW.kind
		elif p >= 2.0:
			# HID++ 2.0 devices do not support registers
			assert self._rw.kind == FeatureRW.kind

		o = _copy(self)
		o._value = None
		o._device = device
		return o

	@property
	def choices(self):
		assert hasattr(self, '_value')
		assert hasattr(self, '_device')

		return self._validator.choices if self._validator.kind & KIND.choice else None

	@property
	def range(self):
		assert hasattr(self, '_value')
		assert hasattr(self, '_device')

		if self._validator.kind == KIND.range:
			return (self._validator.min_value, self._validator.max_value)

	def read(self, cached=True):
		assert hasattr(self, '_value')
		assert hasattr(self, '_device')

		if _log.isEnabledFor(_DEBUG):
			_log.debug("%s: settings read %r from %s", self.name, self._value, self._device)

		if self._value is None and self._device.persister:
			# We haven't read a value from the device yet,
			# maybe we have something in the configuration.
			self._value = self._device.persister.get(self.name)

		if cached and self._value is not None:
			if self._device.persister and self.name not in self._device.persister:
				# If this is a new device (or a new setting for an old device),
				# make sure to save its current value for the next time.
				self._device.persister[self.name] = self._value
			return self._value

		if self._device.online:
			reply = self._rw.read(self._device)
			if reply:
				self._value = self._validator.validate_read(reply)
			if self._device.persister and self.name not in self._device.persister:
				# Don't update the persister if it already has a value,
				# otherwise the first read might overwrite the value we wanted.
				self._device.persister[self.name] = self._value
			return self._value

	def write(self, value):
		assert hasattr(self, '_value')
		assert hasattr(self, '_device')
		assert value is not None

		if _log.isEnabledFor(_DEBUG):
			_log.debug("%s: settings write %r to %s", self.name, value, self._device)

		if self._device.online:
			# Remember the value we're trying to set, even if the write fails.
			# This way even if the device is offline or some other error occurs,
			# the last value we've tried to write is remembered in the configuration.
			self._value = value
			if self._device.persister:
				self._device.persister[self.name] = value

			current_value = None
			if self._validator.needs_current_value:
				# the validator needs the current value, possibly to merge flag values
				current_value = self._rw.read(self._device)

			data_bytes = self._validator.prepare_write(value, current_value)
			if data_bytes is not None:
				if _log.isEnabledFor(_DEBUG):
					_log.debug("%s: settings prepare write(%s) => %r", self.name, value, data_bytes)

				reply = self._rw.write(self._device, data_bytes)
				if not reply:
					# tell whomever is calling that the write failed
					return None

			return value

	def apply(self):
		assert hasattr(self, '_value')
		assert hasattr(self, '_device')

		if _log.isEnabledFor(_DEBUG):
			_log.debug("%s: apply %s (%s)", self.name, self._value, self._device)

		value = self.read()
		if value is not None:
			self.write(value)

	def __str__(self):
		if hasattr(self, '_value'):
			assert hasattr(self, '_device')
			return '<Setting([%s:%s] %s:%s=%s)>' % (self._rw.kind, self._validator.kind, self._device.codename, self.name, self._value)
		return '<Setting([%s:%s] %s)>' % (self._rw.kind, self._validator.kind, self.name)
	__unicode__ = __repr__ = __str__

#
# read/write low-level operators
#

class RegisterRW(object):
	__slots__ = ('register', )

	kind = _NamedInt(0x01, 'register')

	def __init__(self, register):
		assert isinstance(register, int)
		self.register = register

	def read(self, device):
		return device.read_register(self.register)

	def write(self, device, data_bytes):
		return device.write_register(self.register, data_bytes)


class FeatureRW(object):
	__slots__ = ('feature', 'read_fnid', 'write_fnid')

	kind = _NamedInt(0x02, 'feature')
	default_read_fnid = 0x00
	default_write_fnid = 0x10

	def __init__(self, feature, read_fnid=default_read_fnid, write_fnid=default_write_fnid):
		assert isinstance(feature, _NamedInt)
		self.feature = feature
		self.read_fnid = read_fnid
		self.write_fnid = write_fnid

	def read(self, device):
		assert self.feature is not None
		return device.feature_request(self.feature, self.read_fnid)

	def write(self, device, data_bytes):
		assert self.feature is not None
		return device.feature_request(self.feature, self.write_fnid, data_bytes)

#
# value validators
# handle the conversion from read bytes, to setting value, and back
#

class BooleanValidator(object):
	__slots__ = ('true_value', 'false_value', 'mask', 'needs_current_value')

	kind = KIND.toggle
	default_true = 0x01
	default_false = 0x00
	# mask specifies all the affected bits in the value
	default_mask = 0xFF

	def __init__(self, true_value=default_true, false_value=default_false, mask=default_mask):
		if isinstance(true_value, int):
			assert isinstance(false_value, int)
			if mask is None:
				mask = self.default_mask
			else:
				assert isinstance(mask, int)
			assert true_value & false_value == 0
			assert true_value & mask == true_value
			assert false_value & mask == false_value
			self.needs_current_value = (mask != self.default_mask)
		elif isinstance(true_value, bytes):
			if false_value is None or false_value == self.default_false:
				false_value = b'\x00' * len(true_value)
			else:
				assert isinstance(false_value, bytes)
			if mask is None or mask == self.default_mask:
				mask = b'\xFF' * len(true_value)
			else:
				assert isinstance(mask, bytes)
			assert len(mask) == len(true_value) == len(false_value)
			tv = _bytes2int(true_value)
			fv = _bytes2int(false_value)
			mv = _bytes2int(mask)
			assert tv & fv == 0
			assert tv & mv == tv
			assert fv & mv == fv
			self.needs_current_value = any(m != b'\xFF' for m in mask)
		else:
			raise Exception("invalid mask '%r', type %s" % (mask, type(mask)))

		self.true_value = true_value
		self.false_value = false_value
		self.mask = mask

	def validate_read(self, reply_bytes):
		if isinstance(self.mask, int):
			reply_value = ord(reply_bytes[:1]) & self.mask
			if _log.isEnabledFor(_DEBUG):
				_log.debug("BooleanValidator: validate read %r => %02X", reply_bytes, reply_value)
			if reply_value == self.true_value:
				return True
			if reply_value == self.false_value:
				return False
			_log.warn("BooleanValidator: reply %02X mismatched %02X/%02X/%02X",
							reply_value, self.true_value, self.false_value, self.mask)
			return False

		count = len(self.mask)
		mask = _bytes2int(self.mask)
		reply_value = _bytes2int(reply_bytes[:count]) & mask

		true_value = _bytes2int(self.true_value)
		if reply_value == true_value:
			return True

		false_value = _bytes2int(self.false_value)
		if reply_value == false_value:
			return False

		_log.warn("BooleanValidator: reply %r mismatched %r/%r/%r",
						reply_bytes, self.true_value, self.false_value, self.mask)
		return False

	def prepare_write(self, new_value, current_value=None):
		if new_value is None:
			new_value = False
		else:
			assert isinstance(new_value, bool)

		to_write = self.true_value if new_value else self.false_value

		if isinstance(self.mask, int):
			if current_value is not None and self.needs_current_value:
				to_write |= ord(current_value[:1]) & (0xFF ^ self.mask)
			if current_value is not None and to_write == ord(current_value[:1]):
				return None
		else:
			to_write = bytearray(to_write)
			count = len(self.mask)
			for i in range(0, count):
				b = ord(to_write[i:i+1])
				m = ord(self.mask[i : i + 1])
				assert b & m == b
				# b &= m
				if current_value is not None and self.needs_current_value:
					b |= ord(current_value[i : i + 1]) & (0xFF ^ m)
				to_write[i] = b
			to_write = bytes(to_write)

			if current_value is not None and to_write == current_value[:len(to_write)]:
				return None

		if _log.isEnabledFor(_DEBUG):
			_log.debug("BooleanValidator: prepare_write(%s, %s) => %r", new_value, current_value, to_write)

		return to_write


class ChoicesValidator(object):
	__slots__ = ('choices', 'flag', '_bytes_count', 'needs_current_value')

	kind = KIND.choice

	"""Translates between NamedInts and a byte sequence.
	:param choices: a list of NamedInts
	:param bytes_count: the size of the derived byte sequence. If None, it
	will be calculated from the choices."""
	def __init__(self, choices, bytes_count=None):
		assert choices is not None
		assert isinstance(choices, _NamedInts)
		assert len(choices) > 2
		self.choices = choices
		self.needs_current_value = False

		max_bits = max(x.bit_length() for x in choices)
		self._bytes_count = (max_bits // 8) + (1 if max_bits % 8 else 0)
		if bytes_count:
			assert self._bytes_count <= bytes_count
			self._bytes_count = bytes_count
		assert self._bytes_count < 8

	def validate_read(self, reply_bytes):
		reply_value = _bytes2int(reply_bytes[:self._bytes_count])
		valid_value = self.choices[reply_value]
		assert valid_value is not None, "%s: failed to validate read value %02X" % (self.__class__.__name__, reply_value)
		return valid_value

	def prepare_write(self, new_value, current_value=None):
		if new_value is None:
			choice = self.choices[:][0]
		else:
			if isinstance(new_value, int):
				choice = self.choices[new_value]
			elif int(new_value) in self.choices:
				choice = self.choices[int(new_value)]
			elif new_value in self.choices:
				choice = self.choices[new_value]
			else:
				raise ValueError(new_value)

		if choice is None:
			raise ValueError("invalid choice %r" % new_value)
		assert isinstance(choice, _NamedInt)
		return choice.bytes(self._bytes_count)

class RangeValidator(object):
	__slots__ = ('min_value', 'max_value', 'flag', '_bytes_count', 'needs_current_value')

	kind = KIND.range

	"""Translates between integers and a byte sequence.
	:param min_value: minimum accepted value (inclusive)
	:param max_value: maximum accepted value (inclusive)
	:param bytes_count: the size of the derived byte sequence. If None, it
	will be calculated from the range."""
	def __init__(self, min_value, max_value, bytes_count=None):
		assert max_value > min_value
		self.min_value = min_value
		self.max_value = max_value
		self.needs_current_value = False

		self._bytes_count = math.ceil(math.log(max_value + 1, 256))
		if bytes_count:
			assert self._bytes_count <= bytes_count
			self._bytes_count = bytes_count
		assert self._bytes_count < 8

	def validate_read(self, reply_bytes):
		reply_value = _bytes2int(reply_bytes[:self._bytes_count])
		assert reply_value >= self.min_value, "%s: failed to validate read value %02X" % (self.__class__.__name__, reply_value)
		assert reply_value <= self.max_value, "%s: failed to validate read value %02X" % (self.__class__.__name__, reply_value)
		return reply_value

	def prepare_write(self, new_value, current_value=None):
		if new_value < self.min_value or new_value > self.max_value:
			raise ValueError("invalid choice %r" % new_value)
		return _int2bytes(new_value, self._bytes_count)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from gi.repository import Gtk, Gdk

# from logging import getLogger
# _log = getLogger(__name__)
# del getLogger


from solaar.i18n import _

#
#
#

def make(name, label, function, stock_id=None, *args):
	action = Gtk.Action(name, label, label, None)
	action.set_icon_name(name)
	if stock_id is not None:
		action.set_stock_id(stock_id)
	if function:
		action.connect('activate', function, *args)
	return action


def make_toggle(name, label, function, stock_id=None, *args):
	action = Gtk.ToggleAction(name, label, label, None)
	action.set_icon_name(name)
	if stock_id is not None:
		action.set_stock_id(stock_id)
	action.connect('activate', function, *args)
	return action

#
#
#

# def _toggle_notifications(action):
# 	if action.get_active():
# 		notify.init('Solaar')
# 	else:
# 		notify.uninit()
# 	action.set_sensitive(notify.available)
# toggle_notifications = make_toggle('notifications', 'Notifications', _toggle_notifications)


from .about import show_window as _show_about_window
from solaar import NAME
about = make('help-about', _("About") + ' ' + NAME, _show_about_window, stock_id=Gtk.STOCK_ABOUT)

#
#
#

from . import pair_window
def pair(window, receiver):
	assert receiver
	assert receiver.kind is None

	pair_dialog = pair_window.create(receiver)
	pair_dialog.set_transient_for(window)
	pair_dialog.set_destroy_with_parent(True)
	pair_dialog.set_modal(True)
	pair_dialog.set_type_hint(Gdk.WindowTypeHint.DIALOG)
	pair_dialog.set_position(Gtk.WindowPosition.CENTER)
	pair_dialog.present()


from ..ui import error_dialog
def unpair(window, device):
	assert device
	assert device.kind is not None

	qdialog = Gtk.MessageDialog(window, 0,
								Gtk.MessageType.QUESTION, Gtk.ButtonsType.NONE,
								_("Unpair") + ' ' + device.name + ' ?')
	qdialog.set_icon_name('remove')
	qdialog.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
	qdialog.add_button(_("Unpair"), Gtk.ResponseType.ACCEPT)
	choice = qdialog.run()
	qdialog.destroy()
	if choice == Gtk.ResponseType.ACCEPT:
		receiver = device.receiver
		assert receiver
		device_number = device.number

		try:
			del receiver[device_number]
		except:
			# _log.exception("unpairing %s", device)
			error_dialog('unpair', device)



#!/usr/bin/env python3

from glob import glob as _glob
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

autostart_path = '/etc/xdg/autostart'

#from solaar import NAME, __version__
__version__ = '1.0.2.1'
NAME = 'Solaar'


def _data_files():
	from os.path import dirname as _dirname

	yield 'share/solaar/icons', _glob('share/solaar/icons/solaar*.svg')
	yield 'share/solaar/icons', _glob('share/solaar/icons/light_*.png')
	yield 'share/icons/hicolor/scalable/apps', ['share/solaar/icons/solaar.svg']

	for mo in _glob('share/locale/*/LC_MESSAGES/solaar.mo'):
		yield _dirname(mo), [mo]

	yield 'share/applications', ['share/applications/solaar.desktop']
	yield autostart_path, ['share/autostart/solaar.desktop']
	yield '/etc/udev/rules.d', ['rules.d/42-logitech-unify-permissions.rules']

	del _dirname

setup(name=NAME.lower(),
		version=__version__,
		description='Linux devices manager for the Logitech Unifying Receiver.',
		long_description='''
Solaar is a Linux device manager for Logitech's Unifying Receiver peripherals.
It is able to pair/unpair devices with the receiver, for many devices show
battery status, and show and modify some of the modifiable features of devices.
'''.strip(),
		author='Daniel Pavel',
		license='GPLv2',
		url='http://pwr-solaar.github.io/Solaar/',
		classifiers=[
			'Development Status :: 4 - Beta',
			'Environment :: X11 Applications :: GTK',
			'Environment :: Console',
			'Intended Audience :: End Users/Desktop',
			'License :: DFSG approved',
			'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
			'Natural Language :: English',
			'Programming Language :: Python :: 3 :: Only',
			'Operating System :: POSIX :: Linux',
			'Topic :: Utilities',
			],

		platforms=['linux'],

		# sudo apt install python-gi python3-gi \
		#        gir1.2-gtk-3.0 gir1.2-notify-0.7 gir1.2-ayatanaappindicator3-0.1
		# os_requires=['gi.repository.GObject (>= 2.0)', 'gi.repository.Gtk (>= 3.0)'],

		python_requires='>=3.2',
		install_requires=['pyudev (>= 0.13)', ],
		package_dir={'': 'lib'},
		packages=['hidapi', 'logitech_receiver', 'solaar', 'solaar.ui', 'solaar.cli'],
		data_files=list(_data_files()),
		scripts=_glob('bin/*'),
	)
